//-----------------------------------------------------------------------------
//
//  $Logfile:: /game/gamescript.h                                             $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 6:36 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Subclass of script that preprocesses labels
//

#ifndef __GAMESCRIPT_H__
#define __GAMESCRIPT_H__

#include "class.h"
#include "script.h"
#include "archive.h"

typedef struct
	{
	scriptmarker_t pos;
	Burger::String labelname;
	} script_label_t;

class GameScript;

class EXPORT_FROM_DLL GameScriptMarker : public Class
	{
	public:
		CLASS_PROTOTYPE( GameScriptMarker );

      Burger::String               filename;
		scriptmarker_t		scriptmarker;
      virtual void      Archive( Archiver &arc );
      virtual void      Unarchive( Archiver &arc );
	};

inline EXPORT_FROM_DLL void GameScriptMarker::Archive
	(
	Archiver &arc
	)

   {
   // Game scripts are unique in that we don't call our superclass to archive it's data.
   // Instead, we only read enough info to then initialize the script ourselves.
   arc.WriteCString( filename.c_str() );
   arc.WriteBoolean( scriptmarker.tokenready );
   arc.WriteInteger( scriptmarker.offset );
   arc.WriteInteger( scriptmarker.line );
   arc.WriteRaw( scriptmarker.token, sizeof( scriptmarker.token ) );
   }

inline EXPORT_FROM_DLL void GameScriptMarker::Unarchive(Archiver &arc)
{
	// Game scripts are unique in that we don't call our superclass to archive it's data.
	// Instead, we only read enough info to then initialize the script ourselves.
	arc.ReadString( &filename );
	scriptmarker.tokenready = arc.ReadBoolean();
	scriptmarker.offset = arc.ReadInteger();
	scriptmarker.line = arc.ReadInteger();
	arc.ReadRaw( scriptmarker.token, sizeof( scriptmarker.token ) );
}
	

class EXPORT_FROM_DLL GameScript : public Script
	{
	protected:
		Container<script_label_t *>	*labelList;
		GameScript							*sourcescript;
      unsigned                      crc;

	public:
      CLASS_PROTOTYPE( GameScript );

								GameScript();
								GameScript( GameScript *scr );
								~GameScript();
		void					Close( void );
		void					SetSourceScript( GameScript *scr );
		void					LoadFile( const char *filename );

		void					Mark( GameScriptMarker *mark );
		void					Restore( GameScriptMarker *mark );

		void					FreeLabels( void );
		void					FindLabels( void );
		qboolean				labelExists( const char *name );
		qboolean				Goto( const char *name );
      virtual void      Archive( Archiver &arc );
      virtual void      Unarchive( Archiver &arc );
	};


class EXPORT_FROM_DLL ScriptLibrarian : public Class
{
protected:
	Container<GameScript *>	scripts;
	Burger::String                     dialog_script;
	Burger::String                     game_script;

public:
	CLASS_PROTOTYPE( ScriptLibrarian );

	~ScriptLibrarian();

	void			CloseScripts( void );
	void			SetDialogScript( Burger::String scriptname );
	void			SetGameScript( Burger::String scriptname );
	const char  *GetGameScript( void );
	GameScript  *FindScript( const char *name );
	GameScript	*GetScript( const char *name );
	qboolean		Goto( GameScript *scr, const char *name );
	qboolean		labelExists( GameScript *scr, const char *name );
	virtual void Archive( Archiver &arc );
	virtual void Unarchive( Archiver &arc );
};

inline EXPORT_FROM_DLL void ScriptLibrarian::Archive
	(
	Archiver &arc
	)
   {
   GameScript * scr;
   int i, num;

   Class::Archive( arc );

   num = scripts.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      scr = scripts.ObjectAt( i );
      arc.WriteObject( scr );
      }
   arc.WriteCString( dialog_script.c_str() );
   arc.WriteCString( game_script.c_str() );
   }

inline EXPORT_FROM_DLL void ScriptLibrarian::Unarchive
	(
	Archiver &arc
	)
   {
   GameScript * scr;
   int i, num;

   Class::Unarchive( arc );

   scripts.FreeObjectList();

   num = arc.ReadInteger();
   for ( i = 1; i <= num; i++ )
      {
      scr = new GameScript;
      arc.ReadObject( scr );
      scripts.AddObject( scr );
      }

   arc.ReadString( &dialog_script );
   arc.ReadString( &game_script );
   }

extern ScriptLibrarian ScriptLib;

inline EXPORT_FROM_DLL void GameScript::Archive
	(
	Archiver &arc
	)

   {
   // Game scripts are unique in that we don't call our superclass to archive it's data.
   // Instead, we only read enough info to then initialize the script ourselves.
   GameScriptMarker mark;

   arc.WriteUnsigned( crc );

   Mark( &mark );
   arc.WriteObject( &mark );
   }

inline EXPORT_FROM_DLL void GameScript::Unarchive
	(
	Archiver &arc
	)

   {
   // This function is based in part on Restore, so it changes, we must update this function as well.
   // Game scripts are unique in that we don't call our superclass to archive it's data.
   // Instead, we only read enough info to then initialize the script ourselves.
   GameScriptMarker mark;
   unsigned filecrc;
   GameScript *scr;

   filecrc = arc.ReadUnsigned();
   arc.ReadObject( &mark );

   scr = ScriptLib.FindScript( mark.filename.c_str() );
   if ( scr )
      {
      SetSourceScript( scr );
      }
   else
      {
      LoadFile( mark.filename.c_str() );
      }

   // Error out if CRCs have changed
   if ( filecrc != crc )
      {
      gi.error( "File '%s' has changed from when this savegame was written.  Load cancelled.\n", filename.c_str() );
      }

	RestorePosition( &mark.scriptmarker );
   }

#endif
