//-----------------------------------------------------------------------------
//
//  $Logfile:: /game/listener.h                                               $
// $Revision:: 3                                                              $
//   $Author:: Burger                                                         $
//     $Date:: 12/27/00 8:34p                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// 

#include "g_local.h"
#include "class.h"
#include "container.h"

#ifndef __LISTENER_H__
#define __LISTENER_H__

class Entity;
class ScriptVariable;

typedef enum
	{
	EV_FROM_CODE,
	EV_FROM_CONSOLE,
	EV_FROM_SCRIPT
	} eventsource_t;

// Event flags
#define EV_CONSOLE	1								// Allow entry from console
#define EV_CHEAT		2								// Only allow entry from console if cheats are enabled
#define EV_HIDE		4								// Hide from eventlist

#define MAX_EVENT_USE	( ( 1 << 8 ) - 1 )

class ScriptThread;
class Archiver;


class Event : public Class
{
//	private:
public:
	typedef struct EventInfo {
/*		unsigned			inuse			: 8;		// must change MAX_EVENT_USE to reflect maximum number stored here
		unsigned			source		: 2;
		unsigned			flags			: 2;
		unsigned			linenumber	: 20;		// linenumber does double duty in the case of the console commands
*/
		unsigned			linenumber	: 20;		// linenumber does double duty in the case of the console commands
		unsigned			flags			: 2;
		unsigned			source		: 2;
		unsigned			inuse			: 8;		// must change MAX_EVENT_USE to reflect maximum number stored here
	};

	int					eventnum;
	EventInfo			info;
	const char        *name;
	Container<Burger::String>		*data;
	int					threadnum;
	
	Event( int num );
	static void initCommandList( void );
	static void releaseCommandList(void);
	
	friend class		Listener;
	
	friend void	G_ProcessPendingEvents( void );
	friend void G_ClearEventList( void );
	friend void G_InitEvents( void );
	friend void G_ArchiveEvents( Archiver &arc );
	friend void G_UnarchiveEvents( Archiver &arc );
	
	static LinkedList_t commandList;
	static Container<int>	*/*Event::*/flagList;
	static Container<int>	*/*Event::*/sortedList;
	static qboolean			/*Event::*/dirtylist;
	
	static int			compareEvents( const void *arg1, const void *arg2 );
	static void			SortEventList( void );
	static int			FindEvent( const char *name );
	
	public:

		CLASS_PROTOTYPE( Event );

		static int			NumEventCommands( void );
		static void       ListCommands( const char *mask = NULL );

								Event();
								Event(const Event &ev );
								Event( Event *ev );
								Event( const char *command, int flags = -1 );
								~Event();

		Burger::String					getName( void );

		void					SetSource( eventsource_t source ) { info.source = ( unsigned )source; }
		void					SetLineNumber( int linenumber ) { info.linenumber = linenumber; }
		void					SetConsoleEdict( edict_t *consoleedict );
		void					SetThread( ScriptThread *thread );

		eventsource_t		GetSource( void );
		ScriptThread		*GetThread( void );
		edict_t				*GetConsoleEdict( void );
		int					GetLineNumber( void );

		void					Error( const char *fmt, ... );

		static Event		Find( const char *command );
		static qboolean	Exists( const char *command );
		
		Event&				printInfo(void);
		
		operator				int() const;
		operator				const char *();

		int					NumArgs( void );

		qboolean				IsVectorAt( int pos );
		qboolean				IsEntityAt( int pos );
		qboolean				IsNumericAt( int pos );

		const char			*GetToken( int pos );
		const char			*GetString( int pos );
		int					GetInteger( int pos );
		double				GetDouble( int pos );
		float					GetFloat( int pos );
		Vector				GetVector( int pos );
		Entity				*GetEntity( int pos );
		ScriptVariable		*GetVariable( int pos );

		void					AddToken( const char *text ) { AddString( text ); }
		void					AddTokens( int argc, const char **argv );
		void					AddString( const char *text );
		void					AddString( Burger::String &text );
		void					AddInteger( int val );
		void					AddDouble( double val );
		void					AddFloat( float val );
		void					AddVector(const Vector &vec );
		void					AddEntity( Entity *ent );

      virtual void      Archive( Archiver &arc );
      virtual void      Unarchive( Archiver &arc );
	};

extern Event NullEvent;
extern Event EV_Remove;


class Listener;

class EXPORT_FROM_DLL Listener : public Class
	{
	private:
		void							FloatVarEvent( Event *e );
		void							IntVarEvent( Event *e );
		void							StringVarEvent( Event *e );
		void							CharPtrVarEvent( Event *e );
		void							VectorVarEvent( Event *e );
		void							ScriptRemove( Event *e );

	protected:
		void							FloatVar( Event &e, float *var, float defaultvalue = 0 );
		void							IntVar( Event &e, int *var, float defaultvalue = 0 );
		void							StringVar( Event &e, Burger::String *var, const char *defaultvalue = "" );
		void							StringVar( Event &e, char **var, const char *defaultvalue = "" );
		void							VectorVar( Event &e, Vector *var, Vector defaultvalue = Vector( 0, 0, 0 ) );

		qboolean						CheckEventFlags( Event *event );

	public:
		CLASS_PROTOTYPE( Listener );

										~Listener();
		void							Remove( Event *e );
		qboolean						ValidEvent( Event &e );
		qboolean						ValidEvent( const char *name );
		qboolean						EventPending( Event &ev );
		qboolean						ProcessEvent( Event *event );
		qboolean						ProcessEvent(const Event &event );
		void							PostEvent( Event *event, float time );
		void							PostEvent(const Event &event, float time );
		qboolean						PostponeEvent( Event &event, float time );
		qboolean						PostponeEvent( Event *event, float time ) {return PostponeEvent( *event, time );}
		void							CancelEventsOfType( Event *event );
		void							CancelEventsOfType( Event &event ) { CancelEventsOfType( &event ); }
		void							CancelPendingEvents( void );
		qboolean						ProcessPendingEvents( void );
	};


inline eventsource_t Event::GetSource
	(
	void
	)

	{
	return ( eventsource_t )info.source;
	}

inline edict_t *Event::GetConsoleEdict
	(
	void
	)

	{
	// linenumber does double duty in the case of the console commands
	if ( ( info.source != EV_FROM_CONSOLE ) || ( info.linenumber < 1 ) || ( info.linenumber > game.maxclients ) )
		{
		assert( NULL );

		// default to player 1 for release
		return &g_edicts[ 1 ];
		}

	return &g_edicts[ info.linenumber ];
	}

inline EXPORT_FROM_DLL int Event::GetLineNumber
	(
	void
	)

	{
	// linenumber does double duty in the case of the console commands
	if ( info.source == EV_FROM_SCRIPT )
		{
		return static_cast<int>(info.linenumber);
		}

	return 0;
	}

inline EXPORT_FROM_DLL Burger::String Event::getName(void)
{
   assert( name || !eventnum );

   if ( !name )
		{
		return "NULL";
		}

   return name;
	}

inline EXPORT_FROM_DLL Event& Event::printInfo
	(
	void
	)

	{
	gi.dprintf( "event '%s' is number %d\n", getName().c_str(), eventnum );

	return *this;
	}

inline EXPORT_FROM_DLL Event::operator int() const
	{
	return eventnum;
	}

inline EXPORT_FROM_DLL Event::operator const char *()
	{
	return getName().c_str();
	}


#endif
