//-----------------------------------------------------------------------------
//
//  $Logfile:: /GameSource/archive.cpp                                        $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 7:44 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
// 
// DESCRIPTION:
// Class for archiving objects
// 

#include "g_local.h"
#include "archive.h"

#define ARCHIVE_NONE 0
#define ARCHIVE_WRITE 1
#define ARCHIVE_READ  2

#ifdef EXPORT_TEMPLATE
template class EXPORT_FROM_DLL Container<Class *>;
template class EXPORT_FROM_DLL Container<pointer_fixup_t *>;
#endif

enum {
	ARC_NULL, ARC_Vector, ARC_Integer, ARC_Unsigned, ARC_Byte, ARC_Char, ARC_Short, ARC_UnsignedShort,
	ARC_Float, ARC_Double, ARC_Boolean, ARC_String, ARC_Raw, ARC_Object, ARC_ObjectPointer,
	ARC_SafePointer, ARC_Event, ARC_Quat, ARC_Entity,
	ARC_NUMTYPES
};

static const char *typenames[] = {
	"NULL", "vector", "int", "unsigned", "byte", "char", "short", "unsigned short",
	"float", "double", "qboolean", "string", "raw data", "object", "objectpointer",
	"safepointer", "event", "quaternion", "entity"
};

static const char InvalidDataSize[] = "Invalid data size of %d on %s.";
static const char Expecting[] = "Expecting %s";

#define ArchiveHeader   (0x004E4953)		// "SIN"
#define ArchiveVersion	2                       // This must be changed any time the format changes!
#define ArchiveInfo     "Sin Archive Version 2" // This must be changed any time the format changes!

/**********************************

	Handle a file archive for save and load game

**********************************/

CLASS_DECLARATION( Class, Archiver, NULL );

ResponseDef Archiver::Responses[] =
{
	{ NULL, NULL }
};

/**********************************

	Create an archive instance

**********************************/

Archiver::Archiver()
{
	FastMemSet(&fp,0,sizeof(fp));
	numclassespos = 0;
	fileerror = FALSE;
	archivemode = ARCHIVE_NONE;
}

/**********************************

	Dispose of my archive

**********************************/

Archiver::~Archiver()
{
	Close();
	StreamHandleDestroy(&fp);
}

/**********************************

	Crap! Set the error flag and shut myself down

**********************************/

void Archiver::FileError(const char *fmt,...)
{
	va_list	argptr;
	char text[ 1024 ];
	
	va_start( argptr, fmt );
	vsprintf( text, fmt, argptr );
	va_end( argptr );
	
	fileerror = TRUE;
	Close();
	if ( archivemode == ARCHIVE_READ ) {
		gi.error( "Error while loading %s : %s\n", FileName.c_str(), text );
	} else {
		gi.error( "Error while writing to %s : %s\n", FileName.c_str(), text );
	}
}

/**********************************

	Close my input/output file

**********************************/

void Archiver::Close(void)
{
	if ( archivemode == ARCHIVE_WRITE ) {
		Word32 eof;
		FILE *file;
		// write out the number of classpointers
		eof = StreamHandleGetMark(&fp);
		StreamHandleSetMark(&fp,numclassespos);
		WriteInteger( classpointerList.NumObjects() );
		StreamHandleSetMark(&fp,eof);
		StreamHandleEndSave(&fp);
		file = fopen(FileName.c_str(),"wb");
		if (file) {
			fwrite(LockAHandle((void **)fp.DataHandle),1,eof,file);
			UnlockAHandle((void **)fp.DataHandle);
			fclose(file);
		}
	}
	
	StreamHandleDestroy(&fp);
	
	if ( archivemode == ARCHIVE_READ ) {
		int i, num;
		Class * classptr;
		pointer_fixup_t *fixup;
	
		num = fixupList.NumObjects();
		for( i = 1; i <= num; i++ ) {
			fixup = fixupList.ObjectAt( i );
			classptr = classpointerList.ObjectAt( fixup->index );
			if ( fixup->type == pointer_fixup_normal ) {
				Class ** fixupptr;
				fixupptr = ( Class ** )fixup->ptr;
				*fixupptr = classptr;
			} else if ( fixup->type == pointer_fixup_safe ) {
				SafePtrBase * fixupptr;
				fixupptr = ( SafePtrBase * )fixup->ptr;
				fixupptr->InitSafePtr( classptr );
			}
			delete fixup;
		}
		fixupList.FreeObjectList();
		classpointerList.FreeObjectList();
	}
	archivemode = ARCHIVE_NONE;
}

/**********************************

	Open a file for reading

**********************************/

void Archiver::Read(const char *name)
{
	unsigned header;
	unsigned version;
	char *info;
	int num;
	int i;
	Class *FooObj;
	
	Close();
	if ( !name ) {
		gi.error( "NULL pointer for filename in Archiver::Read.\n" );
	}
	
	fileerror = FALSE;
	archivemode = ARCHIVE_READ;
	FileName = name;
	
	if (!gi.LoadFileStream(name,&fp ) ) {
		FileError( "Couldn't open file %s.",name );
	}
	
	header = ReadUnsigned();
	if ( header != ArchiveHeader ) {
		StreamHandleDestroy(&fp);
		FileError( "Not a valid Sin archive." );
	}
	
	version = ReadUnsigned();
	if ( version > ArchiveVersion ) {
		StreamHandleDestroy(&fp);
		FileError( "Archive is from version %.2f.  Check http://www.ritual.com for an update.", version );
	}
	
	if ( version < ArchiveVersion ) {
		StreamHandleDestroy(&fp);
		FileError( "Archive is out of date." );
	}
	
	info = ReadCString();
	gi.dprintf( "%s\n", info);
	Free(info);
	
	// setup out class pointers
	num = ReadInteger();
	classpointerList.Resize( num );
	FooObj = 0;
	for( i = 1; i <= num; i++ ) {
		classpointerList.AddObject(FooObj);
	}
}

/**********************************

	Check if the type is the right data type

**********************************/

inline void Archiver::CheckType(int type)
{
	if (!fileerror) {				/* No errors so far? */
		if (StreamHandleGetLong(&fp) != type ) {		/* Get the data type enum */
			FileError(Expecting, typenames[ type ] );	/* Drat, print the type I was expecting */
		}
	}
}

/**********************************

	Read in the data size
	
**********************************/

inline size_t Archiver::ReadSize(void)
{
	return StreamHandleGetLong(&fp);
}

/**********************************

	Check if the size is the expected size

**********************************/

inline void Archiver::CheckSize(int type,size_t size)
{
	size_t s;
	if (!fileerror) {
		s = StreamHandleGetLong(&fp);
		if (size != s ) {
			FileError( InvalidDataSize, s, typenames[ type ] );
		}
	}
}

/**********************************

	Read raw data

**********************************/

inline void Archiver::ReadData(int type,void *data,size_t size)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != type ) {
			FileError(Expecting, typenames[ type ] );
		}
		s = StreamHandleGetLong(&fp);
		if ( size != s ) {
			FileError(InvalidDataSize, s, typenames[ type ] );
		} else if (size) {
			StreamHandleGetMem(&fp,data,size);
		}
	}
}

/**********************************

	Read an unsigned byte

**********************************/

Word8 Archiver::ReadByte(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Byte ) {
			FileError(Expecting, typenames[ ARC_Byte ] );
		}
		s = StreamHandleGetLong(&fp);
		if (1 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Byte ] );
		} else {
			return StreamHandleGetByte(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a signed byte

**********************************/

char Archiver::ReadChar(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Char ) {
			FileError(Expecting, typenames[ ARC_Char ] );
		}
		s = StreamHandleGetLong(&fp);
		if ( 1 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Char ] );
		} else {
			return (char)StreamHandleGetByte(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a signed short

**********************************/

short Archiver::ReadShort(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Short ) {
			FileError(Expecting, typenames[ ARC_Short ] );
		}
		s = StreamHandleGetLong(&fp);
		if (2 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Short ] );
		} else {
			return (short)StreamHandleGetShort(&fp);
		}
	}
	return 0;
}

/**********************************

	Read an unsigned short

**********************************/

Word16 Archiver::ReadUnsignedShort(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_UnsignedShort ) {
			FileError(Expecting, typenames[ ARC_UnsignedShort ] );
		}
		s = StreamHandleGetLong(&fp);
		if ( 2 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_UnsignedShort ] );
		} else {
			return StreamHandleGetShort(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a signed integer (32 bit)

**********************************/

int Archiver::ReadInteger(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Integer ) {
			FileError(Expecting, typenames[ ARC_Integer ] );
		}
		s = StreamHandleGetLong(&fp);
		if (4 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Integer ] );
		} else {
			return (int)StreamHandleGetLong(&fp);
		}
	}
	return 0;
}

/**********************************

	Read an unsigned short

**********************************/

Word Archiver::ReadUnsigned(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Unsigned ) {
			FileError(Expecting, typenames[ ARC_Unsigned ] );
		}
		s = StreamHandleGetLong(&fp);
		if (4 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Unsigned] );
		} else {
			return (Word)StreamHandleGetLong(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a quake boolean

**********************************/

qboolean Archiver::ReadBoolean(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Boolean ) {
			FileError(Expecting, typenames[ ARC_Boolean ] );
		}
		s = StreamHandleGetLong(&fp);
		if (4 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Boolean] );
		} else {
			return (qboolean)StreamHandleGetLong(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a float

**********************************/

float Archiver::ReadFloat(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Float ) {
			FileError(Expecting, typenames[ ARC_Float ] );
		}
		s = StreamHandleGetLong(&fp);
		if (4 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Float] );
		} else {
			return StreamHandleGetFloat(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a double

**********************************/

double Archiver::ReadDouble(void)
{
	if ( !fileerror ) {
		size_t s;
		if (StreamHandleGetLong(&fp) != ARC_Double ) {
			FileError(Expecting, typenames[ ARC_Double ] );
		}
		s = StreamHandleGetLong(&fp);
		if (8 != s ) {
			FileError(InvalidDataSize, s, typenames[ ARC_Double] );
		} else {
			return StreamHandleGetDouble(&fp);
		}
	}
	return 0;
}

/**********************************

	Read a 3 float vector

**********************************/

void Archiver::ReadVector(Vector *v)
{
	ReadData(ARC_Vector, v, sizeof(Vector));
#if defined(__BIGENDIAN__)
	Burger::SwapEndian::Var(&v->x);
	Burger::SwapEndian::Var(&v->y);
	Burger::SwapEndian::Var(&v->z);
#endif
}

/**********************************

	Read a 4 float quaternion

**********************************/

void Archiver::ReadQuat(Quat *v)
{
	ReadData(ARC_Quat, v, sizeof(Quat));
	
#if defined(__BIGENDIAN__)
	Burger::SwapEndian::Var(&v->x);
	Burger::SwapEndian::Var(&v->y);
	Burger::SwapEndian::Var(&v->z);
	Burger::SwapEndian::Var(&v->w);
#endif
}

void Archiver::ReadObjectPointer(Class ** ptr)	
{
	int index;
	pointer_fixup_t *fixup;
	
	ReadData( ARC_ObjectPointer, &index, sizeof( index ) );
	index = LittleLong(index);
	
	// Check for a NULL pointer
	assert( ptr );
	if ( !ptr ) {
		FileError( "NULL pointer in ReadObjectPointer." );
	}
	
	//
	// see if the variable was NULL
	//
	if ( index == ARCHIVE_NULL_POINTER ) {
		*ptr = NULL;
	} else {
		// init the pointer with NULL until we can fix it
		*ptr = NULL;
	
		fixup = new pointer_fixup_t;
		fixup->ptr = ( void ** )ptr;
		fixup->index = index;
		fixup->type = pointer_fixup_normal;
		fixupList.AddObject( fixup );
	}
}

void Archiver::ReadSafePointer(SafePtrBase * ptr)
{
	int index;
	pointer_fixup_t *fixup;
	
	ReadData( ARC_SafePointer, &index, sizeof( &index ) );
	index = LittleLong(index);
	
	// Check for a NULL pointer
	assert( ptr );
	if ( !ptr ) {
		FileError( "NULL pointer in ReadSafePointer." );
	}
	
	//
	// see if the variable was NULL
	//
	if ( index == ARCHIVE_NULL_POINTER ) {
		ptr->InitSafePtr( NULL );
	} else {
		// init the pointer with NULL until we can fix it
		ptr->InitSafePtr( NULL );
	
		// Add new fixup
		fixup = new pointer_fixup_t;
		fixup->ptr = ( void ** )ptr;
		fixup->index = index;
		fixup->type = pointer_fixup_safe;
		fixupList.AddObject( fixup );
	}
}

void Archiver::ReadEvent(Event * ev)
{
	CheckType( ARC_Event );
	if ( !fileerror ) {
		ev->Unarchive( *this );
	}
}

void Archiver::ReadRaw(void *data,size_t size)
{
	ReadData( ARC_Raw, data, size );
}

char *Archiver::ReadCString(void)
{
	size_t s;
	char *data;

	CheckType( ARC_String );

	if ( !fileerror ) {
		s = StreamHandleGetLong(&fp);
		if ( !fileerror ) {	
			data = (char *)Alloc(s + 1);
			if (data) {
				if (s) {	
					StreamHandleGetMem(&fp,data, s );
				}
				data[s] = 0;
			}
			return data;
		}
	}
	return (char *)AllocClear(1);
}

void Archiver::ReadString(Burger::String *ptr)
{
	size_t	s;
	char *data;

	CheckType( ARC_String );

	if ( !fileerror ) {
		s = StreamHandleGetLong(&fp);
		if ( !fileerror ) {	
			data = (char *)Alloc(s+1);
			if (data) {
				if (s) {	
					StreamHandleGetMem(&fp,data, s );
				}
				data[s] = 0;
				ptr[0] = data;
				Free(data);
			}
			return;
		}
	}
	ptr[0] = "";
}

Class *Archiver::ReadObject(void)
{
	ClassDef	*cls;
	Class		*obj;
	Burger::String		classname;
	long		objstart;
	long		endpos;
	int      index;
	size_t	size;
	qboolean isent;
	int      type;
	
	type = StreamHandleGetLong(&fp);
	if ( ( type != ARC_Object ) && ( type != ARC_Entity ) ) {
		FileError( "Expecting %s or %s", typenames[ ARC_Object ], typenames[ ARC_Entity ] );
	}
	
	size = StreamHandleGetLong(&fp);
	ReadString(&classname);

	cls = getClass( classname.c_str() );
	if ( !cls ) {
		FileError( "Invalid class %s.", classname.c_str() );
	}
	
	isent = checkInheritance( &Entity::ClassInfo, cls );
	if ( type == ARC_Entity ) {
		if ( !isent ) {
			FileError( "Non-Entity class object '%s' saved as an Entity based object.", classname.c_str() );
		}
		game.force_entnum = TRUE;
		game.spawn_entnum = ReadInteger();
	} else if ( isent ) {
		FileError( "Entity class object '%s' saved as non-Entity based object.", classname.c_str() );
	}
	
	index = ReadInteger();
	objstart = StreamHandleGetMark(&fp);
	
	obj = ( Class * )cls->newInstance();
	if ( !obj ) {
		FileError( "Failed to on new instance of class %s.", classname.c_str() );
	} else {
		obj->Unarchive( *this );
	}
	
	if ( isent ) {
		game.force_entnum = FALSE;
	}
	
	if ( !fileerror ) {
		endpos = StreamHandleGetMark(&fp);
		if ( ( endpos - objstart ) > size ) {
			FileError( "Object read past end of object's data" );
		} else if ( ( endpos - objstart ) < size ) {
			FileError( "Object didn't read entire data from file" );
		}
	}
	
	//
	// register this pointer with our list
	//
	classpointerList.AddObjectAt( index, obj );
	
	return obj;
}
	
Class *Archiver::ReadObject(Class *obj)
{
	ClassDef	*cls;
	Burger::String		classname;
	long		objstart;
	long		endpos;
	int      index;
	size_t	size;
	int      type;
	qboolean isent;
	
	type = StreamHandleGetLong(&fp);
	if ( ( type != ARC_Object ) && ( type != ARC_Entity ) ) {
		FileError( "Expecting %s or %s", typenames[ ARC_Object ], typenames[ ARC_Entity ] );
	}
	
	size = StreamHandleGetLong(&fp);
	ReadString(&classname);
	
	cls = getClass( classname.c_str() );
	if ( !cls ) {
		FileError( "Invalid class %s.", classname.c_str() );
	}
	
	if ( obj->classinfo() != cls ) {
		FileError( "Archive has a '%s' object, but was expecting a '%s' object.", classname.c_str(), obj->getClassname() );
	}
	
	isent = obj->isSubclassOf( Entity );
	if ( type == ARC_Entity ) {
		if ( !isent ) {
			FileError( "Non-Entity class object '%s' saved as an Entity based object.", classname.c_str() );
		}
	
		( ( Entity * )obj )->SetEntNum( ReadInteger() );
	} else if ( isent ) {
		FileError( "Entity class object '%s' saved as non-Entity based object.", classname.c_str() );
	}
	
	index = ReadInteger();
	objstart = StreamHandleGetMark(&fp);
	
	obj->Unarchive( *this );
	
	if ( !fileerror ) {
		endpos = StreamHandleGetMark(&fp);
		if ( ( endpos - objstart ) > size ) {
			FileError( "Object read past end of object's data" );
		} else if ( ( endpos - objstart ) < size ) {
			FileError( "Object didn't read entire data from file" );
		}
	}
	
	//
	// register this pointer with our list
	//
	classpointerList.AddObjectAt( index, obj );
	
	return obj;
}
	
/****************************************************************************************

  File Write functions

*****************************************************************************************/

void Archiver::Create(const char *name)
{
	Close();
	if ( !name ) {
		gi.error( "NULL pointer for filename in Archiver::Create.\n" );
	}
	fileerror = FALSE;
	archivemode = ARCHIVE_WRITE;

	FileName = name;

	gi.CreatePath(FileName.c_str());
	StreamHandleInitPut(&fp);
	WriteUnsigned( ArchiveHeader );
	WriteUnsigned( ArchiveVersion );
	WriteCString( ArchiveInfo );
	numclassespos = StreamHandleGetMark(&fp);
	WriteInteger( 0 );
}

inline void Archiver::WriteData(int type,const void *data,size_t size)
{
	StreamHandlePutLong(&fp,type);
	StreamHandlePutLong(&fp,size);
	if ( !fileerror && size ) {
		StreamHandlePutMem(&fp,data, size);
	}
}

#define WRITE( func, type )		\
void Archiver::Write##func(type v) \
{								\
	WriteData( ARC_##func, &v, sizeof( type ) ); \
}

#define WRITESWAP( func, type, swap )									\
void Archiver::Write##func(type v) \
{  \
	type v2 = swap(v);										\
	WriteData( ARC_##func, &v2, sizeof( type ) );			\
}

	WRITESWAP( Integer, int, LittleLong );
	WRITESWAP( Unsigned, unsigned, LittleUnsignedLong );
	WRITE( Byte, byte );
	WRITE( Char, char );
	WRITESWAP( Short, short, LittleShort );
	WRITESWAP( UnsignedShort, unsigned short, LittleUnsignedShort );
	WRITESWAP( Float, float, LittleFloat );
	WRITESWAP( Double, double, LittleDouble );
	WRITESWAP( Boolean, qboolean,LittleLong );


void Archiver::WriteVector(const Vector &v)
{
	Word32 v2[3];
	v2[0] = Burger::LittleEndian::Load(((Word32 *)&v.x));
	v2[1] = Burger::LittleEndian::Load(((Word32 *)&v.y));
	v2[2] = Burger::LittleEndian::Load(((Word32 *)&v.z));
	WriteData( ARC_Vector, &v2,12);
}


void Archiver::WriteQuat(const Quat &v)
{
	Word32 v2[4];
	v2[0] = Burger::LittleEndian::Load(((Word32 *)&v.x));
	v2[1] = Burger::LittleEndian::Load(((Word32 *)&v.y));
	v2[2] = Burger::LittleEndian::Load(((Word32 *)&v.z));
	v2[3] = Burger::LittleEndian::Load(((Word32 *)&v.w));
	WriteData( ARC_Quat, &v2,16);
}


void Archiver::WriteRaw(const void *data,size_t size)
{
	WriteData( ARC_Raw, data, size );
}

void Archiver::WriteCString(const char *string)
{
	WriteData( ARC_String, string,Burger::StringLength(string));
}

void Archiver::WriteObject(Class *obj)
{
	Burger::String		classname;
	long		sizepos;
	long		objstart;
	long		endpos;
	int      index;
	size_t	size;
	qboolean isent;
	
	assert( obj );
	if ( !obj ) {
		FileError( "NULL object in WriteObject" );
	}
	
	isent = obj->isSubclassOf( Entity );
	
	if ( isent ) {
		StreamHandlePutLong(&fp,ARC_Entity);
	} else {
		StreamHandlePutLong(&fp,ARC_Object);
	}
	
	sizepos = StreamHandleGetMark(&fp);
	size = 0;
	StreamHandlePutLong(&fp,size);
	
	classname = obj->getClassname();
	WriteCString( classname.c_str() );
	
	if ( isent ) {
		// Write out the entity number
		WriteInteger( ( ( Entity * )obj )->entnum );
	}
	
	// write out pointer index for this class pointer
	index = classpointerList.AddUniqueObject( obj );
	WriteInteger( index );
	
	if ( !fileerror ) {
		objstart = StreamHandleGetMark(&fp);
		obj->Archive( *this );
	}
	
	if ( !fileerror ) {
		endpos = StreamHandleGetMark(&fp);
		size = endpos - objstart;
		StreamHandleSetMark(&fp,sizepos);
		StreamHandlePutLong(&fp,size);
		StreamHandleSetMark(&fp,endpos);
	}
}

void Archiver::WriteObjectPointer(Class * ptr)
{
	int index;
	
	if ( ptr ) {
		index = classpointerList.AddUniqueObject( ptr );
	} else {
		index = ARCHIVE_NULL_POINTER;
	}
	index = LittleLong(index);
	WriteData( ARC_ObjectPointer, &index, sizeof( index ) );
}

void Archiver::WriteSafePointer(Class * ptr)
{
	int index;

	if ( ptr ) {
		index = classpointerList.AddUniqueObject( ptr );
	} else {
		index = ARCHIVE_NULL_POINTER;
	}
	index = LittleLong(index);
	WriteData( ARC_SafePointer, &index, sizeof( index ) );
}

void Archiver::WriteEvent(Event &ev)
{
	StreamHandlePutLong(&fp,ARC_Event);
	
	//FIXME!!!! Make this handle null events
	if ( &ev == NULL ) {
		NullEvent.Archive( *this );
	} else {
		ev.Archive( *this );
	}
}
