//-----------------------------------------------------------------------------
//
//  $Logfile:: /GameSource/class.h                                            $
// $Revision:: 2                                                              $
//   $Author:: Burger                                                         $
//     $Date:: 12/27/00 8:34p                                                 $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Base class that all classes that are used in conjunction with Sin should
// be based off of.  Class gives run-time type information about any class
// derived from it.  This is really handy when you have a pointer to an object 
// that you need to know if it supports certain behaviour.
// 

#ifndef __CLASS_H__
#define __CLASS_H__

#include "g_local.h"

class Class;
class Event;
class Archiver;

typedef void ( Class::*Response )( Event *event );
typedef struct {
	Event		*event;
	Response	response;
} ResponseDef;

/***********************************************************************

  ClassDef

***********************************************************************/

class ClassDef {
public:
	const char *classname;
	const char *classID;
	const char *superclass;
	void *( *newInstance )( void );
	int	classSize;
	ResponseDef *responses;
	int numEvents;
	Response **responseLookup;
	ClassDef *super;
	ClassDef *next;
	ClassDef *prev;
	
	ClassDef();
	~ClassDef();
	ClassDef( const char *classname, const char *classID, const char *superclass, 
		ResponseDef *responses, void *( *newInstance )( void ), int classSize );
	void	BuildResponseList( void );
};
	
/***********************************************************************

  SafePtr

***********************************************************************/

class SafePtrBase;

class Class;

class SafePtrBase {
private:
	void	AddReference( Class *ptr );
	void	RemoveReference( Class *ptr );
	
protected:
	SafePtrBase *prevSafePtr;
	SafePtrBase *nextSafePtr;
	Class       *ptr;
	
public:
	SafePtrBase();
	virtual     ~SafePtrBase();
	void	      InitSafePtr( Class *newptr );
	void	      Clear( void );
};
	
/***********************************************************************

  Class

***********************************************************************/

#define CLASS_DECLARATION( nameofsuperclass, nameofclass, classid )	\
	ClassDef nameofclass::ClassInfo \
		( \
		#nameofclass, classid, #nameofsuperclass, \
		nameofclass::Responses, nameofclass::_newInstance, \
		sizeof( nameofclass ) \
		); \
	EXPORT_FROM_DLL void *nameofclass::_newInstance( void )	\
	{ \
		return new nameofclass;	\
	} \
	ClassDef *nameofclass::classinfo( void ) \
	{ \
		return &( nameofclass::ClassInfo ); \
	}

#define CLASS_PROTOTYPE( nameofclass ) \
public:	\
	static	ClassDef ClassInfo; \
	static	void *_newInstance( void ); \
	virtual	ClassDef *classinfo( void ); \
	static	ResponseDef Responses[]

class Class {
private:
	SafePtrBase	 *SafePtrList;
	friend class SafePtrBase;
	
public:
	CLASS_PROTOTYPE( Class );
	Class();
	virtual				~Class();
	virtual void		Archive( Archiver &arc );
	virtual void		Unarchive( Archiver &arc );
	void					warning( const char *function, const char *fmt, ... );
	void					error( const char *function, const char *fmt, ... );
	qboolean				inheritsFrom( ClassDef *c );
	qboolean				inheritsFrom( const char *name );
	qboolean				isInheritedBy( const char *name );
	qboolean				isInheritedBy( ClassDef *c );
	const char			*getClassname( void );
	const char			*getClassID( void );
	const char			*getSuperclass( void );
	void					*newInstance( void );
	static void release(void);
	};

void BuildEventResponses( void );
ClassDef *getClassForID( const char *name );
ClassDef *getClass( const char *name );
void listAllClasses( void );
void listInheritanceOrder( const char *classname );
qboolean checkInheritance( ClassDef *superclass, ClassDef *subclass );
qboolean checkInheritance( ClassDef *superclass, const char *subclass );
qboolean checkInheritance( const char *superclass, const char *subclass );
void DisplayMemoryUsage( void );

inline qboolean Class::inheritsFrom(ClassDef *c )
{
	return checkInheritance( c, classinfo() );
}

inline qboolean Class::isInheritedBy(ClassDef *c )
{
	return checkInheritance( classinfo(), c );
}

// The lack of a space between the ")" and "inheritsFrom" is intentional.
// It allows the macro to compile like a function call.  However, this
// may cause problems in some compilers (like gcc), so we may have to
// change this to work like a normal macro with the object passed in
// as a parameter to the macro.
#define isSubclassOf( classname )inheritsFrom( &classname::ClassInfo )
#define isSuperclassOf( classname )isInheritedBy( &classname::ClassInfo )

/***********************************************************************

  SafePtr

***********************************************************************/

inline SafePtrBase::SafePtrBase()
{
	prevSafePtr = NULL;
	nextSafePtr = NULL;
	ptr = NULL;
}

inline SafePtrBase::~SafePtrBase()
{
	Clear();
}

inline void SafePtrBase::Clear(void)
{
	if ( ptr ) {
		RemoveReference( ptr );
		ptr = NULL;
	}
}

inline void SafePtrBase::InitSafePtr(Class *newptr)
{
	if ( ptr != newptr ) {
		if ( ptr ) {
			RemoveReference( ptr );
		}
		ptr = newptr;
		if ( ptr == NULL ) {
			return;
		}
		AddReference( ptr );
	}
}

inline void SafePtrBase::AddReference(Class *ptr)
{
	if ( !ptr->SafePtrList ) {
		ptr->SafePtrList = this;
		LL_Reset( this, nextSafePtr, prevSafePtr );
	} else {
		LL_Add( ptr->SafePtrList, this, nextSafePtr, prevSafePtr );
	}
}

inline void SafePtrBase::RemoveReference(Class *ptr)
{
	if ( ptr->SafePtrList == this ) {
		if ( ptr->SafePtrList->nextSafePtr == this ) {
			ptr->SafePtrList = NULL;
		} else {
			ptr->SafePtrList = nextSafePtr;
			LL_Remove( this, nextSafePtr, prevSafePtr );
		}
	} else {
		LL_Remove( this, nextSafePtr, prevSafePtr );
	}
}

template<class T>
class /*EXPORT_FROM_DLL*/ SafePtr : public SafePtrBase
{
public:
	SafePtr( T* objptr = 0 );
	SafePtr( const SafePtr& obj );
	
	SafePtr& operator=( const SafePtr& obj );
	SafePtr& operator=( T * const obj );
	
	inline int operator==( const T *a ) const { return ptr == a; }
	inline int operator!=( const T *a ) const { return ptr != a; }
	
	inline int operator==( const SafePtr &a ) const { return ptr == a.ptr; }
	inline int operator!=( const SafePtr &a ) const { return ptr != a.ptr; }
	
	operator	T*() const;
	T* operator->() const;
	T& operator*() const;
	T* Ptr(void); //### so I can type cast it when I need to
};

template<class T>
inline SafePtr<T>::SafePtr( T* objptr )
{
	InitSafePtr( objptr );
}

template<class T>
inline SafePtr<T>::SafePtr( const SafePtr& obj )
{
	InitSafePtr( obj.ptr );
}

template<class T>
inline SafePtr<T>& SafePtr<T>::operator=( const SafePtr& obj )
{
	InitSafePtr( obj.ptr );
	return *this;
}

template<class T>
inline SafePtr<T>& SafePtr<T>::operator=( T * const obj )
{
	InitSafePtr( obj );
	return *this;
}

template<class T>
inline int operator==(const T* a,const SafePtr<T> &b)
{
	return b == a;
}

template<class T>
inline int operator!=(const T* a,const SafePtr<T> &b)
{
	return b != a;
}

template<class T, class T2>
inline int operator==(const T* a,const SafePtr<T2> &b)
{
	return b == static_cast<const T2 *>(a);
}

template<class T, class T2>
inline int operator!=(const T* a,const SafePtr<T2> &b)
{
	return b != static_cast<const T2 *>(a);
}


template<class T, class T2>
inline int operator==(const SafePtr<T2> &b,const T* a)
{
	return b == static_cast<const T2 *>(a);
}

template<class T, class T2>
inline int operator!=(const SafePtr<T2> &b,const T* a)
{
	return b != static_cast<const T2 *>(a);
}


template<class T>
inline SafePtr<T>::operator T*() const
{
	return ( T * )ptr;
}

template<class T>
inline T* SafePtr<T>::operator->() const
{
	return ( T * )ptr;
}

template<class T>
inline T& SafePtr<T>::operator*() const
{
	return *( T * )ptr;
}

//### so I can type cast this like a regular pointer
template<class T>
inline T* SafePtr<T>::Ptr(void)
{
	return (T *)ptr;
}

typedef SafePtr<Class> ClassPtr;

#endif /* class.h */
