//-----------------------------------------------------------------------------
//
//  $Logfile:: /GameSource/vector.h                                           $
// $Revision:: 3                                                              $
//   $Author:: Burger                                                         $
//     $Date:: 12/27/00 8:34p                                                 $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// C++ implemention of a Vector object.  Handles standard vector operations
// such as addition, subtraction, normalization, scaling, dot product,
// cross product, length, and decomposition into Euler angles.
// 

#ifndef __VECTOR_H__
#define __VECTOR_H__

#ifndef __BURGER__
#include <burger.h>
#endif

#include "g_local.h"
#include <math.h>
#include <stdio.h>

class Vector : public Vector3D_t {
public:	
	inline Vector() { x = 0; y = 0; z = 0; }
	inline Vector( const vec3_t src ) 	{ x = src[ 0 ]; y = src[ 1 ]; z = src[ 2 ]; }
	inline Vector( const Vector3D_t *src ) { x = src->x; y = src->y; z = src->z; }
	inline Vector( float x1, float y1, float z1 ) { x = x1; y = y1; z = z1; }
	Vector( const char *text );
	
	inline float *vec3( void ) { return &x; }
	inline const float *vec3(void) const { return &x; }
	inline float pitch( void ) const { return x; }
	inline float yaw( void ) const { return y; }
	inline float roll( void ) const { return z; }
	inline float operator[]( int index ) const;
	inline float &operator[]( int index );
	inline void copyTo(vec3_t vec) const { vec[0] = x; vec[ 1 ] = y; vec[ 2 ] = z; }
	inline void setPitch(float pitch ) { x = pitch; }
	inline void setYaw(float yaw) { y = yaw; }
	inline void setRoll(float roll) { z = roll; }
	inline void setXYZ(float x1, float y1, float z1 ) {x = x1; y = y1; z = z1;}
	inline void setZero(void) { x = 0; y = 0; z = 0; }
	inline void operator=(const Vector &a ) { x = a.x; y = a.y; z = a.z; }
	inline void operator=(const vec3_t a ) { x = a[ 0 ]; y = a[ 1 ]; z = a[ 2 ]; }
	inline friend Vector operator+(const Vector &a,const Vector &b) {return Vector(a.x+b.x,a.y+b.y,a.z+b.z);}
	inline Vector& operator+=(const Vector &a ) { x += a.x; y += a.y; z += a.z; return *this; }
	inline friend Vector operator-(const Vector &a,const Vector &b) {return Vector(a.x-b.x,a.y-b.y,a.z-b.z);}
	inline Vector& operator-=(const Vector &a ) { x -= a.x; y -= a.y; z -= a.z; return *this; }
	inline friend Vector operator*(const Vector &a,float b ) {return Vector(a.x*b,a.y*b,a.z*b); }
	inline friend Vector operator*(float a,const Vector &b ) {return Vector(b.x*a,b.y*a,b.z*a); }
	inline friend float operator*(const Vector &a,const Vector &b ) {return a.x * b.x + a.y * b.y + a.z * b.z; }
	inline Vector& operator*=(float a) {x *= a; y *= a; z *= a; return *this; }
	inline Vector& operator/=(float a) {x /= a; y /= a; z /= a; return *this; }
	inline friend int operator==(const Vector &a,const Vector &b) { return ( ( a.x == b.x ) && ( a.y == b.y ) && ( a.z == b.z ) ); }
	inline friend int operator!=(const Vector &a,const Vector &b) { return ( ( a.x != b.x ) || ( a.y != b.y ) || ( a.z != b.z ) ); }
	Vector&	CrossProduct(const Vector &a,const Vector &b);
	float length( void ) const;
	void normalize( void );
	void normalize(const Vector &Input);
	float normalize2( void ); //### returns the length of the original vector
	inline void abs(void) { x = Burger::FPMath::Abs(x); y = Burger::FPMath::Abs(y); z = Burger::FPMath::Abs(z); }
	inline void abs(const Vector &a) { x = Burger::FPMath::Abs(a.x); y = Burger::FPMath::Abs(a.y); z = Burger::FPMath::Abs(a.z); }
	float toYaw(void);
	float toPitch(void);
	void toAngles(const Vector &a);	
	void AngleVectors(Vector *forward, Vector *right, Vector *up ) const;
	void LerpVector(const Vector &w1,const Vector &w2,float t);
	float MaxValue(void) const;
};

extern Vector vec_zero;		/* Should be const, but too many warnings... */

inline float Vector::operator[](int index ) const
{
	assert( ( index >= 0 ) && ( index < 3 ) );
	return ( &x )[ index ];
}

inline float& Vector::operator[](int index)
{
	assert( ( index >= 0 ) && ( index < 3 ) );
	return ( &x )[ index ];
}

inline Vector& Vector::CrossProduct(const Vector &a,const Vector &b)
{
	float x1,y1,z1;
	x1 = a.y * b.z - a.z * b.y;
	y1 = a.z * b.x - a.x * b.z;
	z1 = a.x * b.y - a.y * b.x;
	x = x1;			/* Store AFTER I am done so I can store to myself */
	y = y1;
	z = z1;
	return *this;
}

inline float Vector::length(void) const
{
	float	length;
	
	length = x * x + y * y + z * z;
	return ( float )Burger::FPMath::Sqrt( length );
}

inline void Vector::normalize(void)
{
	float length, ilength;
	
	length = this->length();
	if ( length ) {
		ilength = 1 / length;
		x *= ilength;
		y *= ilength;
		z *= ilength;
	}
}

inline void Vector::normalize(const Vector &Input)
{
	float length, ilength;
	
	length = Input.length();
	if ( length ) {
		ilength = 1 / length;
		x = Input.x*ilength;
		y = Input.y*ilength;
		z = Input.z*ilength;
	} else {
		x = 0;
		y = 0;
		z = 0;
	}
}
//###
// normalize function that returns the length of the original vector
inline float Vector::normalize2 (void)
{
	float	length, ilength;

	length = this->length();
	if (length ) {
		ilength = 1 / length;
		x *= ilength;
		y *= ilength;
		z *= ilength;
	}
	return length;
}
//###

inline float Vector::MaxValue(void) const
{
	float maxy;
	float maxz;
	float maxx;
	
	maxx = Burger::FPMath::Abs( (float)x );
	maxy = Burger::FPMath::Abs( (float)y );
	maxz = Burger::FPMath::Abs( (float)z );
	if ( maxy > maxx )
		maxx = maxy;
	if ( maxz > maxx )
		maxx = maxz;
	return maxx;
}

class Quat {
public:
	float x;
	float y;
	float z;
	float w;
	
	inline Quat() { x = 0; y = 0; z = 0; w = 0; }
	inline Quat(const Vector &Angles ) { EulerToQuat( Angles.vec3(), this->vec4() ); }
	inline Quat(const float srcMatrix[ 3 ][ 3 ] ) { MatToQuat( srcMatrix, this->vec4() ); }
	inline Quat(float x1,float y1,float z1,float w1) { x = x1; y = y1; z = z1; w = w1; }
	
	inline float *vec4(void) { return &x; }
	inline const float *vec4(void) const { return &x; }
	float operator[]( int index ) const;
	float& operator[]( int index );
	inline void set( float x1, float y1, float z1, float w1 ) {x = x1; y = y1;z = z1;w = w1;}
	inline void operator=(const Quat &a ) {x = a.x; y = a.y; z = a.z; w = a.w; }
	inline friend Quat operator+(const Quat &a,const Quat &b) { return Quat( a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w ); }
	inline Quat& operator+=(const Quat &a ) { x += a.x; y += a.y; z += a.z; w += a.w; return *this; }
	inline friend Quat operator-(const Quat &a,const Quat &b) { return Quat( a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w ); }
	inline Quat& operator-=(const Quat &a) { x -= a.x; y -= a.y; z -= a.z; w -= a.w; return *this; }
	inline friend Quat operator*(const Quat &a, float b ) { return Quat( a.x * b, a.y * b, a.z * b, a.w * b ); }
	inline friend Quat operator*( float a,const Quat &b ) { return Quat( b.x * a, b.y * a, b.z * a, b.w * a ); }
	inline Quat& operator*=( float a ) { x *= a; y *= a; z *= a; w *= a; return *this; }
	friend int operator==(const Quat &a,const Quat &b );
	friend int	operator!=(const Quat &a,const Quat &b );
	float length( void );
	Quat& normalize( void );
	Quat operator-();
	Vector toAngles( void );	
};

inline float Quat::operator[](int index ) const
{
	assert( ( index >= 0 ) && ( index < 4 ) );
	return ( &x )[ index ];
}

inline float& Quat::operator[](int index)
{
	assert( ( index >= 0 ) && ( index < 4 ) );
	return ( &x )[ index ];
}

inline int operator==(const Quat &a,const Quat &b)
{
	return ( ( a.x == b.x ) && ( a.y == b.y ) && ( a.z == b.z ) && ( a.w == b.w ) );
}

inline int operator!=(const	Quat &a,const Quat &b)
{
	return ( ( a.x != b.x ) || ( a.y != b.y ) || ( a.z != b.z ) && ( a.w != b.w ) );
}

inline float Quat::length(void)
{
	float	length;
	
	length = x * x + y * y + z * z + w * w;
	return ( float )Burger::FPMath::Sqrt( length );
}

inline Quat& Quat::normalize(void)
{
	float	length, ilength;

	length = this->length();
	if ( length ) {
		ilength = 1 / length;
		x *= ilength;
		y *= ilength;
		z *= ilength;
		w *= ilength;
	}
	return *this;
}

inline Quat Quat::operator-()
{
	return Quat( -x, -y, -z, -w );
}

inline Vector Quat::toAngles(void)
{
	float m[ 3 ][ 3 ];
	vec3_t angles;

	QuatToMat( this->vec4(), m );
	MatrixToEulerAngles( m, angles );
	return Vector( angles );
}


#endif /* Vector.h */
