/*
	MAC_APPLEEVENT.C
	
	This module isolates all the Apple Event handling code for Mac Q2.  
	
	Q2 pays attention to Apple Events in the following ways:
	
		* The 'Run Application' Apple Event has an optional 'params' argument defined in the
		  Quake 2 application aete Apple Event needs resource, this params argument consists of
		  a text string that is interpreted as command line arguments
		  
		* There is a 'console' Apple Event that may be sent to Quake 2, its argument consists
		  of a text string that is interpreted as a console command list
		  
		* Looking into Open Document support for making Quake 2 do things when you open a specific
		  data file type...
		 
		* Implement Quit application command...
*/

#include "macfuncs.h"
#include <AppleEvents.h>

// Quake class ID identifying us

enum {
	kQuakeEventClass = FOUR_CHAR_CODE('Quak')
};

// 'Run' Apple Event command line option param key

enum {
	kQuakeCommandLineParam	=	FOUR_CHAR_CODE('CLin')
};

// Quake console command event IDs

enum {
	kQuakeConsoleCmd	=	FOUR_CHAR_CODE('Comd')
};

/* Globals */
static AEEventHandlerUPP gAppleEventHandler;			// universal proc to the Apple Event handler

/* ------------------------------------------
	AppleEventsGotRequiredParams
	
	Validate we have an Apple Event with all the
	required parameters being passed to us
*/
static Bool AppleEventsGotRequiredParams(const AppleEvent *inEvent)
{
	DescType returnedType;
	Size actualSize;
	OSErr theErr;

	/* Check to see if any required attributes are missing from this Apple Event sent to us */
	theErr = AEGetAttributePtr(inEvent, keyMissedKeywordAttr, typeWildCard, &returnedType, nil, 0,  &actualSize);
	if (errAEDescNotFound == theErr) {
		return TRUE;
	}
	return FALSE;
}

static OSErr GotRequiredParams(const AppleEvent *theAppleEvent)
{
	DescType returnedType;
	Size actualSize;
	OSErr theErr;

	theErr = AEGetAttributePtr(theAppleEvent, keyMissedKeywordAttr,typeWildCard, &returnedType, nil, 0, &actualSize);
	if (theErr == errAEDescNotFound) {
		return noErr;
	}
	if (!theErr) {
		return errAEParamMissed;
	}
	return theErr;
}

//		AppleEventHandler

static OSErr pascal AppleEventHandler(const AppleEvent *inEvent, AppleEvent *outReply,long inRefCon)
{
#pragma unused (outReply, inRefCon)

	OSErr		theErr;
	DescType	actualType;
	Size		actualSize;
	DescType	eventClass, eventID;
	char		theStr[1024];

	/* get the 'class' of Apple Event sent to us */
	
	theErr = AEGetAttributePtr(inEvent, keyEventClassAttr, typeType, &actualType, &eventClass, sizeof (eventClass), &actualSize);
	if (theErr) {
		return theErr;
	}
	/* get the 'id' of the Apple Event sent to us */

	theErr = AEGetAttributePtr(inEvent, keyEventIDAttr, typeType, &actualType, &eventID, sizeof (eventID), &actualSize);
	if (theErr) {
		return theErr;
	}

	/* is the class a Core Event class? (the basic 4 apple events, (run, open, print, quit)) */ 

	if (eventClass == kCoreEventClass) {
		switch (eventID) {
		/* An 'Open Application event */
		case kAEOpenApplication:
			return (noErr);
				
		/* An 'Open Documents' event */
		case kAEOpenDocuments:
			{
				AEDescList docList;
				OSErr theErr;
				long index, itemsInList;
				Size actualSize;
				AEKeyword keywd;
				DescType returnedType;
				FSSpec theFSS;
				
				// In response to this event, we must open all documents specified in
				// the Apple Event.

				AEGetParamDesc(inEvent, keyDirectObject, typeAEList, &docList);
				
				// check for missing required parameters
				theErr = GotRequiredParams(inEvent);
				if (theErr) {
					return AEDisposeDesc(&docList);
				}				
				// count the number of descriptor records in the list
				theErr = AECountItems(&docList, &itemsInList);
				
				// get each descriptor record from the list, coerce the returned
				// data to an FSSpec record, and open the associated file
				index = 1;
				if (index <= itemsInList) {
					char *FileName;
					AEGetNthPtr(&docList, index, typeFSS, &keywd, &returnedType,(Ptr)&theFSS, sizeof(theFSS), &actualSize);
					// open FSSpec
					FileName = Burger::Mac::GetPathFromFSSpec(NULL,0,&theFSS);
					if (FileName) {
						Word32 Type;
						Type = Burger::FileManager::GetFileType(FileName);
						if (Type=='TEXT') {
							WordPtr Length;
							void *Data;
							Data = Burger::FileManager::LoadFile(FileName,&Length);
							if (Data) {
								if (Length>1023) {		/* Failsafe! */
									Length = 1023;
								}
								FastMemCpy(theStr,Data,Length);
								theStr[Length]=0;
								Burger::Free(Data);
								Mac_SetCommandLine(theStr);	/* pass the arguments string on to be parsed */
							}
						}
						Free(FileName);
					}
				}
				AEDisposeDesc(&docList);
			}
			return noErr;
					
		/* A 'Print Documents' event */
		case kAEPrintDocuments:
			return (errAEEventNotHandled);		
				
		/* A 'Quit Application' event */
		case kAEQuitApplication:
			if (AppleEventsGotRequiredParams(inEvent)) {
				return noErr;
			}
			return errAEParamMissed;
							
		default:
			return errAEHandlerNotFound;
		}		
	}
	
	return (noErr);
}

/* ------------------------------------------
	Mac_InitAppleEventHandler
	
	Hook in our Apple Event handler
*/
void Mac_InitAppleEventHandler(void)
{
	/* Create an Apple Event handler that will see all Apple Events sent to us */
	gAppleEventHandler = NewAEEventHandlerUPP(AppleEventHandler);
	if (gAppleEventHandler) {
		AEInstallEventHandler(typeWildCard, typeWildCard, gAppleEventHandler, 0L, FALSE);
	}
}

/* ------------------------------------------
	Mac_KillAppleEventHandler
	
	Take out our Apple Event handler
*/
void Mac_KillAppleEventHandler(void)
{
	/* Zap it */
	if (gAppleEventHandler) {
		AERemoveEventHandler(typeWildCard, typeWildCard, gAppleEventHandler, FALSE);
		DisposeAEEventHandlerUPP(gAppleEventHandler);
		gAppleEventHandler = 0;
	}
}
