// sys_null.h -- null system driver to aid porting efforts

#include <burger.h>
#include "qcommon.h"
#include "rw_mac.h"
#include "errno.h"
#include "mac_libutil.h"
#include "console3d.h"
#include "alias.h"
#include "mac_export.h"
#include "mac_isp.h"
#include "sinmdl.h"
#include "mac_version.h"
#include "server.h"
#include "keys.h"
#include "macfuncs.h"
#include "client.h"
#include <MacMemory.h>

#if defined(__GAMERANGER__)
#include <GameRanger.h>
#endif

#define MAX_NUM_ARGVS 128

/*****************************************************************************/
/* Extern protos/vars                                                        */
/*****************************************************************************/

extern long WriteCharsToConsole(char *buffer, long n);
extern long ReadCharsFromConsole(char *buffer, long n);

extern void MacOSStartup(void);
extern void InitTheBurger(void);
extern void MacOSShutdown(void);

/*****************************************************************************/
/* Misc. globals for file                                                    */
/*****************************************************************************/

void *game_library = 0; 
unsigned long sys_frame_time;	//NCT was just unsigned sys_frame_time

int gMacBackground = 0;		//NCT is the game in 'background' mode so we can get at Mac desktop?
int LaunchingGameRanger;

// For our faux command line interface

#define MAX_CMDL_ARGS	50
char *mac_argv[MAX_CMDL_ARGS];
char mac_clbufferp[2048];
int	mac_argc;

/****************************************************************************************/
/*
	void Sys_Error (char *error, ...)

	Blah, bad Sys_Error happened, make everything die
*/

void Sys_Error (const char *error, ...)
{
	va_list		argptr;
	char errMes[1024];

	CL_Shutdown();				/* shutdown client */
	Qcommon_Shutdown();			/* shutdown common */

	va_start (argptr,error);
	vsprintf(errMes,error,argptr);
	va_end (argptr);
	Debug::Message("Sys_Error: %s\n",errMes);
	DoMacFatalDialog(errMes);

	exit (1);
}

/****************************************************************************************/
/*
	void Sys_Quit (void)
	
	Cleanup and blow this taco stand
*/


void Sys_Quit (void)
{
	CL_Shutdown();
	Qcommon_Shutdown();
	MacOSShutdown();
	SV_ShutdownGameProgs();	/* Dispose of the game dll */
	MENU_ReleasePlayerConfig();		/* Release the skin names */
	MENU_Release();		/* Release level names */
	Cvar_Destroy();		/* Dispose of the vars */
	HUD_Destroy();		/* Release the HUD file */
	FS_Release();		/* Release the pak files */
	SINMDL_FreeAll();	/* Release all the models */
	AliasRelease();		/* Kill the alias manager */
	CM_Release();		/* Release the CModels */
	CON_Destroy();		/* Release the consoles */
	Key_ReleaseBindings();	/* Release the bound keys */
	Cmd_ReleaseStrings();	/* Release the command strings */
#if defined(__GAMERANGER__)
	if (LaunchingGameRanger) {
		GROpenGameRanger();
	}
#endif
	exit(0);
}

/****************************************************************************************/
/*
	void	Sys_UnloadGame (void)

	Unload any currently loaded game library file (gameppc.lib)
*/

void	Sys_UnloadGame (void)
{
	if (game_library != 0)
	{
#ifndef REF_HARD_LINKED
		FreeMacLibrary(game_library);
#endif
		game_library = 0;
	}
}

/****************************************************************************************/
/*
	void	*Sys_GetGameAPI (void *parms)
	
	Load a new game library file (gameppc.lib), need to use the 'game' directory
	cvar to get the base directory name in the Quake directory to use to decide
	where to look for the gameppc.lib file
*/

void	*Sys_GetGameAPI (void *parms)
{
	ImportMacFunctions_t ImportMacFunctions;
	void *(*GetGameAPI)(void *);
	pascal OSErr (*__initialize)(const void *theInitBlock);

	if (game_library)
		Sys_Error(ERR_FATAL,"Sys_GetGameAPI without Sys_UnloadingGame");

	// Load the library
	
	game_library = LoadMacLibrary("gameppc.lib");
	if (game_library == 0) {
		DebugString("Sys_GetGameAPI: couldn't load lib\n");
		return(0);
	}

	// Export our Mac functions to the library

	ImportMacFunctions = (ImportMacFunctions_t)GetSymbol(game_library, "ImportMacFunctions");
	if (ImportMacFunctions == 0)
		Sys_Error(ERR_FATAL, "GetSymbol failed for ImportMacFunctions\n");

	Mac_InitMacExport(ImportMacFunctions());	//export our mac routines to lib

	// Export/import the game API to the library

	__initialize = (pascal OSErr (*)(const void *))GetSymbol(game_library, "__initialize");
	if (__initialize == 0)
		Sys_Error(ERR_FATAL, "__initialize failed for GetGameAPI\n");
	__initialize(0);
	
	GetGameAPI = (void *(*)(void *))GetSymbol(game_library, "GetGameAPI");
	if (GetGameAPI == 0)
		Sys_Error(ERR_FATAL, "GetSymbol failed for GetGameAPI\n");
	return((void *)GetGameAPI(parms));
}

/****************************************************************************************/
/*
	void Sys_InitConsole (void)

*/
void Sys_InitConsole(void)
{
	if (!dedicated || !dedicated->value)
		return;

// If a dedicated console, we don't need InputSprocket active!!!

	InputSetState(FALSE);		/* Shut off InputSprocket */

#if !defined(__STDC__) && 0
	SIOUXSettings.initializeTB = TRUE;		/* do initialize toolbox */
#endif
											/* else we'll crash! */
	Mac_InitInputLine();											
//	SIOUXSettings.standalone = TRUE;		/* Not a standalone SIOUX app */

}


/****************************************************************************************/
/*
	char *Sys_ConsoleInput (void)
*/

char MyConsoleInput[256];
int MyConsoleIOffset = 0;

const char *Sys_ConsoleInput (void)
{
	if (!dedicated || !dedicated->value)
		return(0);

	Mac_ConsoleInput();
	if (MyConsoleIOffset == 0)
		return(NULL);
	MyConsoleIOffset = 0;
	return(MyConsoleInput);

}

/****************************************************************************************/
/*
	void	Sys_ConsoleOutput (char *string)

*/

void	Sys_ConsoleOutput (const char *string)
{
	if (!dedicated || !dedicated->value)
		return;

//	WriteCharsToConsole(string,Burger::StringLength(string));
	printf(string);
}

/****************************************************************************************/
/*
	void Sys_SendKeyEvents (void)
*/

void Sys_SendKeyEvents (void)
{
	if (KBD_Update_fp)
		KBD_Update_fp();
	sys_frame_time = Sys_Milliseconds();
}

/****************************************************************************************/
/*
	void Sys_AppActivate (void)
*/

void Sys_AppActivate (void)
{
}

/****************************************************************************************/
/*
	void Sys_CopyProtect (void)
*/

void Sys_CopyProtect (void)
{
}

/****************************************************************************************/
/*
	char *Sys_GetClipboardData( void )
*/

char *Sys_GetClipboardData( void )
{
	return Mac_GetTextScrap();
}


/****************************************************************************************/
/*
	void Mac_SetCommandLine(char *clstr)

	We've got a big long string of stuff, turn it into traditional command line
	parms for processing by Quake
*/

void Mac_SetCommandLine(char *clstr)
{
	int offset;
	int tracingWhite;	//are tracing whitespace?
	int tracingQuote;

	if (!clstr || !clstr[0]) {
		return;
	}
		
	// we reset/overwrite any existing arguments every time we're called
	tracingWhite = FALSE;
	tracingQuote = FALSE;
	offset = 0;
	mac_argc = 1;
	mac_argv[0] = &(mac_clbufferp[0]);
	
	while (clstr[offset] != 0) {
		mac_clbufferp[offset] = clstr[offset];
		
		if (clstr[offset] == '"') {
			if (tracingQuote == TRUE) {
				tracingQuote = FALSE;
			} else {
				tracingWhite = FALSE;
				tracingQuote = TRUE;
				mac_argv[mac_argc] = &(mac_clbufferp[offset]);
				++mac_argc;
			}
		}
		
		if (tracingQuote == FALSE) {
			if (clstr[offset] == ' ') {
				tracingWhite = TRUE;
			} else {
				if (tracingWhite == TRUE) {
					mac_clbufferp[offset-1] = 0;	// chop it!
					mac_argv[mac_argc] = &(mac_clbufferp[offset]);
					++mac_argc;
					tracingWhite = FALSE;
				}
			}
		}
		++offset;
	}
}

/* --------------------------------------------------------------------
	ScanForCD
	
	Checks to see if a 'Quake2' volume exists and if it does if
	the path to the PC quake2.exe app file is there, if so we have
	our Quake 2 cd-rom drive mounted!
*/
static char *ScanForCD(void)
{
static char cddir[MAX_OSPATH];
static qboolean done;

#ifndef DEMO
	FILE	*f;
	char	test[MAX_QPATH];

	if (done) {
		return(cddir);	//don't recheck
	}
	done = TRUE;
	
	sprintf(cddir, "Sin:SinInstallData");
	sprintf(test, "Sin:SinInstallData:license.txt");
	f = fopen(test, "r");
	if (f) {
		fclose(f);
		return(cddir);
	}
#endif
	cddir[0] = 0;
	return(0);
}

static void KillTools(void)
{
	CL_PredictRelease();
	SV_ClipRelease();
#if defined(_DEBUG)
	{
		extern void Z_Stats_f2(void);
		Z_Stats_f2();
	}
#endif
	MouseDestroy();
	KeyboardDestroy();
	OpenGLRelease();
	ReleaseVideo();
}

//=============================================================================

void main(int argc, char **argv)
{
	int time, oldtime, newtime;
	char *cddir;
	int oktostart;

	/* Set the Metrowerks SIOUX console options*/
	
#if !defined(__STDC__) && 0
	SIOUXSettings.autocloseonquit = FALSE;	/* autoclose SIOUX window on term. */
	SIOUXSettings.asktosaveonclose = FALSE;	/* don't offer to save output */
	SIOUXSettings.initializeTB = FALSE;		/* don't initialize toolbox */
	SIOUXSettings.setupmenus = FALSE;		/* don't setup SIOUX menus */
	SIOUXSettings.standalone = FALSE;		/* Not a standalone SIOUX app */
#endif
	mac_argc = 0;						// Init our faux command line option settings
	mac_argv[0] = 0;					// which may get filled by a pending Apple Event

#if defined(_DEBUG)
	Debug::SetTraceFlag(Debug::TRACE_MEMORYLEAKS); /* |DEBUGTRACE_THEWORKS */;
#endif
	MacOSStartup();						// moved this out of Sys_Init for Apple Event setup reasons...
	atexit(KillTools);
	
	oktostart = CheckOSCompatibility();	// Check for conflicts (extensions and crap)
	if (oktostart) {
		// Setup the cddir for the CD-ROM data path
		cddir = ScanForCD();
		
		if (cddir && mac_argc < MAX_NUM_ARGVS - 3) {
			int i;
			// don't override a cddir on command line
			for (i=0;i<mac_argc;i++)
				if (!strcmp(mac_argv[i], "cddir"))
					break;
			if (i == mac_argc) {
				mac_argv[mac_argc++] = "+set";
				mac_argv[mac_argc++] = "cddir";
				mac_argv[mac_argc++] = cddir;
			}
		}
		
		Qcommon_Init (mac_argc, mac_argv);	// NCT(argc, argv);

#if !defined(__STDC__) && 0
		SIOUXSetTitle("\pSin Dedicated Console");
#endif

		oldtime = Sys_Milliseconds();
		while (1) {
			do {
				newtime = Sys_Milliseconds();
				time = newtime - oldtime;
				//Qcommon_Frame (0.1); bah
			} while (time < 1);

			KeyboardCallPollingProcs();	//NCT_ADD for redbook looping
			Qcommon_Frame(time);
			oldtime = newtime;

			// Do we have to do the Mac background thing?
			if (gMacBackground == TRUE) {
				if (MacOSBackground() == TRUE) {
					Mac_Resume_f();					// we were told not to background anymore
				}
			}
		}
	} else {
		MacOSShutdown();
	}
}


